class DffError(Exception):
    pass


class LammpsError(Exception):
    pass


class PackmolError(Exception):
    pass


class GmxError(Exception):
    pass


class GaussError(Exception):
    pass


class OpenBabelError(Exception):
    pass


class RDKitError(Exception):
    pass


class SchedulerError(Exception):
    pass
